@@ -23,6 +23,7 @@ import socket
 import sys
 import textwrap
 import time
+import inspect
 
 
 MAXFD = 1024
@@ -103,6 +104,8 @@ relative import to an absolute import.
         return sys.modules[name]
 
 def load_class(uri, default="sync", section="gunicorn.workers"):
+    if inspect.isclass(uri):
+        return uri
     if uri.startswith("egg:"):
         # uses entry points
         entry_str = uri.split("egg:")[1]
@@ -122,8 +125,9 @@ def load_class(uri, default="sync", section="gunicorn.workers"):
 
                 return pkg_resources.load_entry_point("gunicorn",
                             section, uri)
-            except ImportError:
-                raise RuntimeError("class uri invalid or not found")
+            except ImportError, e:
+                raise RuntimeError("class uri invalid or not found: " +
+                        "[%s]" % str(e))
         klass = components.pop(-1)
         mod = __import__('.'.join(components))
         for comp in components[1:]:
@@ -340,7 +344,7 @@ def seed():
     try:
         random.seed(os.urandom(64))
     except NotImplementedError:
-        random.seed(random.random())
+        random.seed('%s.%s' % (time.time(), os.getpid()))
 
 
 def check_is_writeable(path):
