@@ -62,10 +62,12 @@ class GeventWorker(AsyncWorker):
             server = StreamServer(self.socket, handle=self.handle, spawn=pool)
 
         server.start()
+        pid = os.getpid()
         try:
             while self.alive:
                 self.notify()
-                if self.ppid != os.getppid():
+
+                if  pid == os.getpid() and self.ppid != os.getppid():
                     self.log.info("Parent changed, shutting down: %s", self)
                     break
 
@@ -77,7 +79,7 @@ class GeventWorker(AsyncWorker):
         try:
             # Try to stop connections until timeout
             self.notify()
-            server.stop(timeout=self.timeout)
+            server.stop(timeout=self.cfg.graceful_timeout)
         except:
             pass
 
@@ -92,6 +94,7 @@ class GeventWorker(AsyncWorker):
         def init_process(self):
             #gevent 0.13 and older doesn't reinitialize dns for us after forking
             #here's the workaround
+            import gevent.core
             gevent.core.dns_shutdown(fail_requests=1)
             gevent.core.dns_init()
             super(GeventWorker, self).init_process()
